#include "HVCore.h"
#include <fstream>

// static keys init
HVCore* HVCore::instance = NULL;
int		HVCore::factor = 0;
bool	HVCore::ctrl = false;


HVCore::HVCore(bool tracking, bool useServer, bool saveCapture)
{
	// init params
	this->tracking = tracking;
	this->useServer = useServer;
	this->saveCapture = saveCapture;

	// default settings
	HVCore::instance = this;
	this->dual = false;
	this->stereoCalibration = false;
	this->calibrate = false;
	this->calRect = cvRect(0, 0, CAM_W, CAM_H);
	this->distBottomOffset = 0;
	this->distTopOffset = 0;
	this->showCenter = false;
	this->showImages = 4;
	this->drawCalibrationLines = false;
	this->autoTracking = 0;
	this->overlayLevel = 1;

	// init VideoInput
	vi = new videoInput();
	vi->setUseCallback(true);
	
	// check for cameras
	int numDev = vi->listDevices(true);
	if (numDev < 1) { printf("### No cameras connected to the computer\n"); return; }
	else if (numDev < 2) { printf("### Only one camera connected. Depth information won't be calculated.\n"); }
	else { dual = true; }

	// init cams
	cam[0] = new HVCamera(vi, 0, &distTopOffset, &distBottomOffset);
	if (dual) 
	{
		cam[1] = new HVCamera(vi, 1, &distTopOffset, &distBottomOffset);
		sview = new StereoView();
	}

	// print info
	printf(	"----------------------------------------------------\n"
			"HVCore v0.0925 by Georg Graf\n"
			"----------------------------------------------------\n"
			"keys:\n"
			"\t 0-3 \t\t set overlay level\n"
			"\t o,p \t\t camera preferences\n"
			"\t s \t\t toggle tracking\n"
			"\t r \t\t reset tracking\n"
			"\t x \t\t create tracking rectangle\n"
			"\t c \t\t start chessboard calibration\n"
			"\t u,i \t\t inc/dec disparity position\n"
			"\t n,m \t\t inc/dec disparity range\n"
			"\t q,w \t\t inc/dec maximum finger radius\n"
			"\t a \t\t toggle amplify\n"
			"\t f \t\t toggle finger recognition\n"
			"\t b \t\t capture background image complex\n"
			"\t v \t\t capture background image simple\n"
			//"\t t,z \t\t image distortion-top\n"
			//"\t v,b \t\t image distortion-bottom\n"
			"\t h \t\t show center of image\n"
			"\t ctrl+s\t\t save settings\n"
			"\t ctrl+l\t\t load settings\n"
			"\t ctrl+r\t\t reset settings\n"
			"\t ctrl+(1-5) \t set displayed images\n"
			//"\n"
			//"\t use shift+key to increase by 5\n"
			"----------------------------------------------------\n");
	
	printf("HVCORE: capture_size: %ix%i, display_size: %ix%i\n", CAM_W, CAM_H, CAM_WIN_W, CAM_WIN_H);

	// init edit img
	tmpImg = cvCreateImage(cvSize(CAM_W, CAM_H), 8, 3);

	// init Window Output
	wout = new WindowOutput("hv11"); // , x, y
	glfwSetKeyCallback(HVCore::keydown);

	// init tracking
	hand = new HandVuLite();
	if (tracking)
	{
		char conductorPath[100];
		sprintf(conductorPath, HV11_CONDUCTOR_FILENAME, CAM_W);
		hand->initLite(conductorPath, 3);
	}

	// init gesture recoginition
	gesture = new Gesture(100 / DISP_CAM_FACTOR);

	// create server
	if (useServer)
		fi = new FI(6055, this);

	if (saveCapture)
		ci = new CI(6088);
	
	// load settings
	if (HV11_LOAD_SETTINGS_ON_STARTUP)
		loadSettings();
}

HVCore::~HVCore(void)
{
	delete fi;

	delete cam[0];
	if (dual) delete cam[1];

	delete sview;
	delete wout;

	delete vi;
}

void HVCore::process(void)
{
	wout->calcFps();
	run();
	track();
	draw();
}

void HVCore::run()
{
	cam[0]->run();
	
	if (dual) 
	{
		cam[1]->run();

		if (stereoCalibration)  
		{	// chessboard calibration
			stereoCalibration = !(sview->handleCalibration(cam[0]->image, cam[1]->image));
		}
		else
		{	// calibrate and calc disparity
			sview->calibrateImages(cam[0]->image, cam[1]->image);
			sview->calcDisparity(cam[0]->image, cam[1]->image);
			//sview->highpassImage(cam[0]->image);
			sview->calcDepth((int)hand->getCenterX(),(int)hand->getCenterY());
		}

		//gesture->calcContours(cam[0]->image);
	}

	// save captured image
	if (saveCapture)
		ci->sendCapture(cam[0]->image, calRect);
}

void HVCore::track(void)
{
	if (tracking)
	{
		// Tracking
		hand->processHv(cam[0]->image, NULL, sview->depth);

		float x = hand->getCenterX();
		float y = hand->getCenterY();

		// Get Gesture
		//if (dual)
		//	gesture->calcGesture(cvPoint((int)x, (int)y), sview->depth, overlayLevel);

		// calibrieren
		if (calibrate)
		{
			if (x < calRect.x) calRect.x = (int)x;
			if (y < calRect.y) calRect.y = (int)y;
			if (x > calRect.x + calRect.width) calRect.width = (int)(x - calRect.x);
			if (y > calRect.y + calRect.height) calRect.height = (int)(y - calRect.y);

			// draw red rect
			cvRectangle(hand->procImage, cvPoint(calRect.x, calRect.y), cvPoint(calRect.x + calRect.width, calRect.y + calRect.height), CV_RGB(0, 0, 255));
		}
		else // draw blue rect
			cvRectangle(hand->procImage, cvPoint(calRect.x, calRect.y), cvPoint(calRect.x + calRect.width, calRect.y + calRect.height), CV_RGB(255, 0, 0));
		
		// send data
		if (useServer)
			fi->send(	((x - float(calRect.x)) / calRect.width),
						((y - float(calRect.y)) / calRect.height),
						sview->depth,
						hand->getFingerState());
	}
}

void HVCore::draw(void)
{
	cam[0]->draw();

	if (dual)
	{
		// output addtitional Window title
		sprintf(wout->additionalTitle, "(depth: %f) (state: %s)", sview->depth, hand->getFingerState());
		
		cam[1]->draw();

		// draw calibration lines
		if (drawCalibrationLines) 
		{
			sview->drawCalibrationLines(cam[0]->image);
			sview->drawCalibrationLines(cam[1]->image);
		}

		sview->draw(overlayLevel);

		switch(showImages)
		{
			case 1: wout->process(hand->procImage); break;
			case 2: wout->process(cam[0]->image, cam[1]->image); break;
			case 3: wout->process(cam[0]->image, cam[1]->image, sview->disparityImage); break;
			case 4: wout->process(cam[0]->image, cam[1]->image, sview->disparityImage, hand->procImage); break;
			case 5: wout->process(cam[0]->image, cam[1]->image, sview->disparityImage, hand->procImage, gesture->gestureImage); break;
		}
	}
	else
	{
		if (showImages > 1) wout->process(cam[0]->image);
		else				wout->process(cam[0]->image, hand->procImage);
	}
	
}

void HVCore::doCommand(char* cmd)
{
	if (tracking)
	{
		if		(strcmp(cmd, CMD_START) == 0)	{ hand->StartRecognition(); }
		else if (strcmp(cmd, CMD_STOP) == 0)	{ hand->StopRecognition(); }
		else if (strcmp(cmd, CMD_TOGGLE) == 0)	{ if (hand->IsActive()) hand->StopRecognition(); else hand->StartRecognition(); }
		else if (strcmp(cmd, CMD_RESET) == 0)	{ hand->StopRecognition(); hand->StartRecognition(); }
	}
}

void HVCore::keydown(int key, int action)
{
	if (action != GLFW_PRESS)
		return;

	factor = (glfwGetKey(GLFW_KEY_LSHIFT) == 1) ? 5 : 1;
	ctrl = (glfwGetKey(GLFW_KEY_LCTRL) == 1);

	if (ctrl)
	{
		switch(key)
		{
			case 'S': instance->saveSettings(); break;
			case 'L': instance->loadSettings(); break;
			case 'R': instance->resetSettings(); break;
			case '1': instance->showImages = 1; break;
			case '2': instance->showImages = 2; break;
			case '3': instance->showImages = 3; break;
			case '4': instance->showImages = 4; break;
			case '5': instance->showImages = 5; break;
		}
	}
	else
	{
		switch(key)
		{
			case 'X': if (!instance->calibrate) instance->calRect = cvRect((int)(instance->hand->getCenterX()), (int)(instance->hand->getCenterY()), 0, 0);
					  instance->calibrate = !instance->calibrate; break;
			case 'S': instance->doCommand(CMD_TOGGLE); break;
			case 'O': instance->vi->showSettingsWindow(0); break;
			case 'P': instance->vi->showSettingsWindow(1); break;
			case 'R': instance->doCommand(CMD_RESET); break;
			case 'L': instance->drawCalibrationLines = !instance->drawCalibrationLines; break;
			case 'C': if (instance->dual) instance->stereoCalibration = !instance->stereoCalibration; 
					  if (instance->stereoCalibration){ printf("HVCore: Chessboard-Calibration started. Waiting for input ...\n"); instance->sview->resetAnalyse(); } 
					  else								printf("\nHVCore: Chessboard-Calibration canceled.\n"); break;
			case 'U': if (instance->sview) instance->sview->setDispValues(instance->sview->getDispPos()-factor, instance->sview->getDispRange()); break;
			case 'I': if (instance->sview) instance->sview->setDispValues(instance->sview->getDispPos()+factor, instance->sview->getDispRange()); break;
			case 'N': if (instance->sview) instance->sview->setDispValues(instance->sview->getDispPos(),		instance->sview->getDispRange()-factor); break;
			case 'M': if (instance->sview) instance->sview->setDispValues(instance->sview->getDispPos(),		instance->sview->getDispRange()+factor); break;
			case 'B': instance->cam[0]->setBg(); if (instance->dual) { instance->cam[1]->setBg(); } break;
			case 'V': instance->cam[0]->setBg(false); if (instance->dual) { instance->cam[1]->setBg(false); } break;
			//case 'T': instance->distTopOffset += factor;break;
			//case 'Z': instance->distTopOffset -= factor; break;
			//case 'V': instance->distBottomOffset += factor;  break;
			//case 'B': instance->distBottomOffset -= factor; break;
			case 'H': instance->showCenter = !instance->showCenter; 
					  instance->cam[0]->showCenter = instance->showCenter; 
					  if (instance->dual) instance->cam[1]->showCenter = instance->showCenter; break;
		}
		if (instance->tracking)
		{
			switch(key)
			{
				case '0': instance->hand->SetOverlayLevel(0); instance->overlayLevel = 0; break;
				case '1': instance->hand->SetOverlayLevel(1); instance->overlayLevel = 1; break;
				case '2': instance->hand->SetOverlayLevel(2); instance->overlayLevel = 2; break;
				case '3': instance->hand->SetOverlayLevel(3); instance->overlayLevel = 3; break;
				case 'Q': instance->hand->maxRadius -= factor; instance->hand->initFingers(); break;
				case 'W': instance->hand->maxRadius += factor; instance->hand->initFingers(); break;
				case 'A': instance->hand->amplifyEnabled = !instance->hand->amplifyEnabled; break;
				case 'F': instance->hand->fingerDetectionEnabled = !instance->hand->fingerDetectionEnabled; break;
			}
		}
	}
}




// ############################################
// ################# SETTINGS #################
// ############################################
void HVCore::saveSettings(void)
{
	printf("HVCORE: save settings ...");
	
	if (sview)
	{
		char cal_file[100];
		sprintf(cal_file, HV11_CAL_FILENAME, CAM_W);

		if (sview->saveCalibration(cal_file))
			printf(" %s", cal_file);
	}

	char settings_file[100];
	sprintf(settings_file, HV11_SETTINGS_FILENAME, CAM_W);
	
	ofstream myfile(settings_file);
	if (myfile.is_open())
	{
		myfile << "#hv11 settings file v0.3" << std::endl;
		myfile << std::endl;
		myfile << "#auto-start tracking, amplify, fingerdetection (1 = true, 0 = false)" << std::endl;
		myfile << "autostart=" << autoTracking << "," << hand->amplifyEnabled << "," << hand->fingerDetectionEnabled << std::endl;
		myfile << "#calibration-rectangle(x, y, w, h)" << std::endl;
		myfile << "calibration=" << calRect.x << "," << calRect.y << "," << calRect.width << "," << calRect.height << std::endl;
		myfile << "#image distortion (top, bottom)" << std::endl;
		myfile << "distortion=" << distTopOffset << "," << distBottomOffset << std::endl;
		myfile << "#overlay level (0..3)" << std::endl;
		myfile << "overlay=" << overlayLevel << std::endl;
		myfile << "#show windows(0..5)" << std::endl;
		myfile << "wincount=" << showImages << std::endl;
		myfile << "#finger properties (max radius, xoffset, yoffset)" << std::endl;
		myfile << "fingerprop=" << hand->maxRadius << "," << hand->xoffset << "," << hand->yoffset << "," << hand->transValue << std::endl;
		myfile << "#amplilfy value (1..*)" << std::endl;
		myfile << "amplify=" << hand->amplifyValue << std::endl;
		if (dual) 
		{
			myfile << "# DISPARITY OPTIONS: range: the distance between near (depth=1) and far (depth=0) as an multiplier " << std::endl;
			myfile << "# for internal values; position: the starting position as equally weighted multiplier" << std::endl;
			myfile << "#disparity-position(int value)" << std::endl;
			myfile << "disparitypos=" << sview->getDispPos() << std::endl;
			myfile << "#disparity-range(unsigned int value > 0)" << std::endl;
			myfile << "disparityrange=" << sview->getDispRange() << std::endl;
		}
		myfile.close();

		printf(" %s done!\n", settings_file);
	}
	else
		printf(" settings couldn't be saved!\n");
}

void HVCore::loadSettings(void)
{
	printf("HVCORE: load settings ...");

	// load calibration, if exists
	if (sview)
	{
		char cal_file[100];
		sprintf(cal_file, HV11_CAL_FILENAME, CAM_W);

		if (sview->loadCalibration(cal_file))
			printf(" %s", cal_file);
	}

	char settings_file[100];
	sprintf(settings_file, HV11_SETTINGS_FILENAME, CAM_W);

	string line;
	ifstream myfile(settings_file);

	if (myfile.is_open())
	{
		int dispPos;
		unsigned int dispRange;

		while(!myfile.eof())
		{
			getline(myfile, line);

			if (line[0] == '#' || line.empty()) continue;
			else
			{
				if (line.find("autostart") != string::npos)			sscanf_s(line.c_str(), "autostart=%i,%i,%i", &autoTracking, &hand->amplifyEnabled, &hand->fingerDetectionEnabled);
				else if (line.find("calibration") != string::npos)	sscanf_s(line.c_str(), "calibration=%i,%i,%i,%i", &calRect.x, &calRect.y, &calRect.width, &calRect.height);
				else if (line.find("distortion") != string::npos)	sscanf_s(line.c_str(), "distortion=%f,%f", &distTopOffset, &distBottomOffset);
				else if (line.find("overlay") != string::npos)		sscanf_s(line.c_str(), "overlay=%i", &overlayLevel);
				else if (line.find("wincount") != string::npos)		sscanf_s(line.c_str(), "wincount=%i", &showImages);
				else if (line.find("gtransition") != string::npos)	sscanf_s(line.c_str(), "gtransition=%i,%i", &gesture->pointTrans, &gesture->spreadTrans);
				else if (line.find("fingerprop") != string::npos)	sscanf_s(line.c_str(), "fingerprop=%i,%i,%i,%f", &hand->maxRadius, &hand->xoffset, &hand->yoffset, &hand->transValue);
				else if (line.find("amplify") != string::npos)		sscanf_s(line.c_str(), "amplify=%f", &hand->amplifyValue);

				if (dual)
				{
					if (line.find("disparitypos") != string::npos)			sscanf_s(line.c_str(), "disparitypos=%i", &dispPos);
					else if (line.find("disparityrange") != string::npos)	sscanf_s(line.c_str(), "disparityrange=%i", &dispRange);
				}
			}
		}

		printf(" %s done!\n", settings_file);
		
		if (autoTracking == 1 && hand)
			hand->StartRecognition();

		if (dispPos && dispRange && sview)
			sview->setDispValues(dispPos, dispRange);

		if (tracking)
		{
			hand->SetOverlayLevel(overlayLevel);
			hand->initFingers();
		}
	}
	else
		printf(" settings couldn't be loaded!\n");
}
void HVCore::resetSettings(void)
{
	printf("HVCORE: reset settings ...");

	calRect.x = 0;
	calRect.y = 0;
	calRect.width = CAM_W;
	calRect.height = CAM_H;

	distTopOffset = 0;
	distBottomOffset = 0;

	if (sview)
		sview->resetAnalyse();

	if (tracking)
	{
		hand->StopRecognition();
		hand->maxRadius = 105;
	}

	printf(" done!\n");
}