#include "StereoView.h"

/**
Some pieces of the calibration code is derived from
the author //starlino//
-> http://www.starlino.com/opencv_qt_stereovision.html
**/

#include <sstream>
#include <algorithm>
#include <ctype.h>

StereoView::StereoView(int numBoards, int waitFramesForNextBoard, int cornersX, int cornersY)
{
	// ### Chessboard params ###
	this->numBoards = numBoards;
	this->waitFrames = waitFramesForNextBoard;
	this->cornersX = cornersX;
	this->cornersY = cornersY;
	this->nCorners = (cornersX * cornersY);
	this->boardSize = cvSize(cornersX, cornersY);
	this->imageSize = cvSize(CAM_W, CAM_H);
	this->depth = 0;
	this->disparityRange = 2;
	this->disparityPosition = -2;
	this->depthRect = cvRect(0, 0, 0, 0);

	// init calibration
	resetAnalyse();

	// ### create disparty image vars ###
	tmpImage = cvCreateImage(cvSize(CAM_DISPARITY_W, CAM_DISPARITY_H), IPL_DEPTH_16S, 1);
	tmpGrayImage = cvCreateImage(cvSize(CAM_W, CAM_H), 8, 1);
	leftTmpImg = cvCreateImage(cvSize(CAM_DISPARITY_W, CAM_DISPARITY_H), 8, 1);
	rightTmpImg = cvCreateImage(cvSize(CAM_DISPARITY_W, CAM_DISPARITY_H), 8, 1);
	disparityImage = cvCreateImage(cvSize(CAM_DISPARITY_W, CAM_DISPARITY_H), 8, 1);

	// ### create calibration tmp vars ###
	lGrayTmpImage = cvCreateImage(cvSize(CAM_W, CAM_H), 8, 1);
	rGrayTmpImage = cvCreateImage(cvSize(CAM_W, CAM_H), 8, 1);

	// ### create highpass tmp vars ###
	highpassedImage = cvCreateImage(cvSize(CAM_W, CAM_H), 8, 3);
	hpTmpImg = cvCreateImage(cvSize(CAM_W, CAM_H), 8, 1);
	
	state = cvCreateStereoBMState();
	setDispValues(disparityPosition, disparityRange);
}

StereoView::~StereoView(void)
{
	cvReleaseImage(&tmpImage);
	cvReleaseImage(&leftTmpImg);
	cvReleaseImage(&rightTmpImg);
	cvReleaseImage(&lGrayTmpImage);
	cvReleaseImage(&rGrayTmpImage);
	cvReleaseImage(&disparityImage);
	cvReleaseStereoBMState(&state);
}

int		StereoView::getDispPos(void) { return disparityPosition; }
unsigned int StereoView::getDispRange(void) { return disparityRange; }
void	StereoView::setDispValues(int pos, unsigned int range) 
{
	if (range < 1)
		return;

	state->preFilterSize = (CAM_W == 640) ? 35 : 25;
    state->preFilterCap = 31;
    state->SADWindowSize = (CAM_W == 640) ? 35 : 25;
    state->textureThreshold = 10;
    state->uniquenessRatio = 15;
	state->minDisparity = pos * 16;
	state->numberOfDisparities = range * 16;

	this->disparityPosition = pos;
	this->disparityRange = range;

	printf("SView:  disparityPosition: %i, disparityRange: %i\n", pos, range);
}


void StereoView::calcDisparity(IplImage* lImg, IplImage* rImg)
{
	cvZero(disparityImage);

	if (lImg->nChannels > 1)
	{
		cvCvtColor(lImg, tmpGrayImage, CV_RGB2GRAY);
		cvResize(tmpGrayImage, leftTmpImg);

		cvCvtColor(rImg, tmpGrayImage, CV_RGB2GRAY);
		cvResize(tmpGrayImage, rightTmpImg);
	}
	else
	{
		cvResize(lImg, leftTmpImg);
		cvResize(rImg, rightTmpImg);
	}

	cvFindStereoCorrespondenceBM(leftTmpImg, rightTmpImg, tmpImage, state);
	cvNormalize(tmpImage, disparityImage, 0, 256, CV_MINMAX);

	//cvFindStereoCorrespondence(lImg, rImg, CV_DISPARITY_BIRCHFIELD, disparityImage, 250, 30, 3, 2, 3, 4);
}

void StereoView::drawCalibrationLines(IplImage* img, int count)
{
	for (int i = 0; i < count; i++)
		cvLine(img, cvPoint(0, ((int)(img->height / count) * i)), cvPoint(img->width, ((int)(img->height / count) * i)), CV_RGB(0, 0, 255));
}

void StereoView::calcDepth(int centerX, int centerY, int size)
{
	unsigned char *di = reinterpret_cast<unsigned char*>(disparityImage->imageData);
	int diStep = disparityImage->widthStep;
	float count = 1;
	float sum = 0.f;
	
	int x = (int)((centerX - size/2) / (float)DISP_CAM_FACTOR);
	int y = (int)((centerY - size/2) / (float)DISP_CAM_FACTOR);

	// border min/max values
	if (x < 0) x = 0;
	else if (x + size > CAM_DISPARITY_W) x = CAM_DISPARITY_W - size;
	if (y < 0) y = 0;
	else if (y + size > CAM_DISPARITY_H) y = CAM_DISPARITY_H - size;
	
	// set start value
	di += diStep * y;

	for (int i = 1; i < size; i++) //height
	{
		for (int j=x; j < x + size; j+= disparityImage->nChannels) //width
		{
			if (di[(int)j] > 5)
			{
				sum += (di[(int)j]);
				count++;
			}
		}
		di += diStep;
	}

	depth = (sum / count) / 255;

	// save to draw
	depthRect.x = x;
	depthRect.y = y;
	depthRect.width = size;
	depthRect.height = size;

	//depth = cvAvg(disparityImage).val[0] / 255;
}

void StereoView::highpassImage(IplImage* img)
{
	cvResize(disparityImage, hpTmpImg, CV_INTER_NN);

	RgbImage hpImg(highpassedImage);
	BwImage dispImg(hpTmpImg);

	if (img->nChannels > 1)
	{
		RgbImage imgImg(img);

		for (int i = 0; i < img->height; i++)
			for (int j = 0; j < img->width; j++)
			{
				if (dispImg(i,j) < HIGHPASS_MIN_VAL)
				{
					hpImg(i,j, 0) = 0;
					hpImg(i,j, 1) = 0;
					hpImg(i,j, 2) = 0;
				}
				else
				{
					hpImg(i,j, 0) = imgImg(i,j, 0);
					hpImg(i,j, 1) = imgImg(i,j, 1);
					hpImg(i,j, 2) = imgImg(i,j, 2);
				}
			}
	}
	else
	{
		BwImage imgImg(img);

		for (int i = 0; i < img->height; i++)
			for (int j = 0; j < img->width; j++)
				if (dispImg(i,j) < HIGHPASS_MIN_VAL)	hpImg(i,j, 0) = hpImg(i,j, 1) = hpImg(i,j, 2) = 0;
				else									hpImg(i,j, 0) = hpImg(i,j, 1) = hpImg(i,j, 2) = imgImg(i,j);			
	}

	//cvMul(hpTwo, hpOne, img);
}

void StereoView::draw(int overlayLevel)
{
	if (overlayLevel > 1)
		cvRectangle(disparityImage, cvPoint(depthRect.x, depthRect.y), cvPoint(depthRect.x + depthRect.width, depthRect.y + depthRect.height), CV_RGB(255, 255, 255));
}


// ############################################################
// ################# CHESSBAORD CALIBRATION ###################
// ############################################################
bool StereoView::loadCalibration(char* filename)
{
    cvReleaseMat(&mx1);
    cvReleaseMat(&my1);
    cvReleaseMat(&mx2);
    cvReleaseMat(&my2);
	mx1 = cvCreateMat(imageSize.height, imageSize.width, CV_32F);
    my1 = cvCreateMat(imageSize.height, imageSize.width, CV_32F);
    mx2 = cvCreateMat(imageSize.height, imageSize.width, CV_32F);
    my2 = cvCreateMat(imageSize.height, imageSize.width, CV_32F);
    FILE* f =  fopen(filename, "rb");
    if(!f) return false;
    if(!fread(mx1->data.fl,sizeof(float), mx1->rows*mx1->cols, f)) return false;
    if(!fread(my1->data.fl,sizeof(float), my1->rows*my1->cols, f)) return false;
    if(!fread(mx2->data.fl,sizeof(float), mx2->rows*mx2->cols, f)) return false;
    if(!fread(my2->data.fl,sizeof(float), my2->rows*my2->cols, f)) return false;
    fclose(f);
    ready = true;
    return true;
}
bool StereoView::saveCalibration(char* filename)
{
    if(!ready) return false;
    FILE* f =  fopen(filename, "wb");
    if(!f) return false;
    if(!fwrite(mx1->data.fl,sizeof(float), mx1->rows*mx1->cols, f)) return false;
    if(!fwrite(my1->data.fl,sizeof(float), my1->rows*my1->cols, f)) return false;
    if(!fwrite(mx2->data.fl,sizeof(float), mx2->rows*mx2->cols, f)) return false;
    if(!fwrite(my2->data.fl,sizeof(float), my2->rows*my2->cols, f)) return false;
    fclose(f);
    return true;
}

void StereoView::resetAnalyse(void)
{
	successes = 0;
	frame = 1;
	ready = false;

	mx1 = my1 = mx2 = my2 = 0;
    pointsTemp[0].resize(nCorners);
    pointsTemp[1].resize(nCorners);
}

bool StereoView::handleCalibration(IplImage *lImg, IplImage *rImg)
{
	if (successes < numBoards)
	{
		if(frame % waitFrames == 0)	
			analyseImages(lImg, rImg);
		else
			if (frame % 5 == 0)
				printf(".");

		frame++;
		return false;
	}
	else // ### all chessboards were found ###
	{
		createRectification();
		printf("### Chessboard recognition finished. camera will be calibrated!\n");
		ready = true;

		return true;
	}
}

bool StereoView::analyseImages(IplImage *lImg, IplImage *rImg)
{
	printf(" testing\n");

	int found = 0 ;
	IplImage* image[2];

	if (lImg->nChannels > 1)
	{
		cvCvtColor(lImg, leftTmpImg, CV_RGB2GRAY);
		cvCvtColor(rImg, rightTmpImg, CV_RGB2GRAY);
		image[0] = leftTmpImg;
		image[1] = rightTmpImg;
	}
	else
	{
		image[0] = lImg;
		image[1] = rImg;
	}
	

	for(int lr=0;lr<2;lr++)
	{
		CvSize imageSize =  cvGetSize(image[lr]);

		int cornersDetected = 0;

		//FIND CHESSBOARDS AND CORNERS THEREIN:
		int result = cvFindChessboardCorners(image[lr], boardSize, &pointsTemp[lr][0], &cornersDetected, CV_CALIB_CB_ADAPTIVE_THRESH | CV_CALIB_CB_NORMALIZE_IMAGE);

		if(result && cornersDetected == nCorners)
		{
			cvFindCornerSubPix(image[lr], &pointsTemp[lr][0], cornersDetected, cvSize(11, 11), cvSize(-1,-1), cvTermCriteria(CV_TERMCRIT_ITER+CV_TERMCRIT_EPS,30, 0.01));
			found++;
		}

	}

	if(found == 2)
	{
		for(int lr=0; lr < 2; lr++)
		{
			points[lr].resize((successes+1)*nCorners);
			copy(pointsTemp[lr].begin(), pointsTemp[lr].end(),  points[lr].begin() + successes * nCorners);
		}
		successes++;
		printf("\n### Chessboard %i/%i found!\n", successes, numBoards);
		return true;
	}
	else
		return false;
}

void StereoView::createRectification(void)
{
	// ARRAY AND VECTOR STORAGE:
	double M1[3][3], M2[3][3], D1[5], D2[5];
	double R[3][3], T[3], E[3][3], F[3][3];
	CvMat _M1,_M2,_D1,_D2,_R,_T,_E,_F;

	_M1 = cvMat(3, 3, CV_64F, M1 );
	_M2 = cvMat(3, 3, CV_64F, M2 );
	_D1 = cvMat(1, 5, CV_64F, D1 );
	_D2 = cvMat(1, 5, CV_64F, D2 );
	_R = cvMat(3, 3, CV_64F, R );
	_T = cvMat(3, 1, CV_64F, T );
	_E = cvMat(3, 3, CV_64F, E );
	_F = cvMat(3, 3, CV_64F, F );

	// HARVEST CHESSBOARD 3D OBJECT POINT LIST:
	objectPoints.resize(successes * nCorners);

	for(int k=0;k<successes;k++)
		for(int i = 0; i < cornersY; i++ )
			for(int j = 0; j < cornersX; j++ )
				objectPoints[k * cornersY * cornersX + i * cornersX + j] = cvPoint3D32f(i, j, 0);

	npoints.resize(successes, nCorners);

	int N = successes * nCorners;

	CvMat _objectPoints = cvMat(1, N, CV_32FC3, &objectPoints[0]);
	CvMat _imagePoints1 = cvMat(1, N, CV_32FC2, &points[0][0]);
	CvMat _imagePoints2 = cvMat(1, N, CV_32FC2, &points[1][0]);
	CvMat _npoints = cvMat(1, npoints.size(), CV_32S, &npoints[0]);
	cvSetIdentity(&_M1);
	cvSetIdentity(&_M2);
	cvZero(&_D1);
	cvZero(&_D2);

	//CALIBRATE THE STEREO CAMERAS
	cvStereoCalibrate(&_objectPoints, &_imagePoints1, &_imagePoints2, &_npoints, &_M1, &_D1, &_M2, &_D2, imageSize, &_R, &_T, &_E, &_F, cvTermCriteria(CV_TERMCRIT_ITER+CV_TERMCRIT_EPS, 100, 1e-5), CV_CALIB_FIX_ASPECT_RATIO + CV_CALIB_ZERO_TANGENT_DIST + CV_CALIB_SAME_FOCAL_LENGTH);

	//Always work in undistorted space
	cvUndistortPoints(&_imagePoints1, &_imagePoints1,&_M1, &_D1, 0, &_M1);
	cvUndistortPoints(&_imagePoints2, &_imagePoints2,&_M2, &_D2, 0, &_M2);

	//COMPUTE AND DISPLAY RECTIFICATION
	double R1[3][3], R2[3][3];
	CvMat _R1 = cvMat(3, 3, CV_64F, R1);
	CvMat _R2 = cvMat(3, 3, CV_64F, R2);

	//HARTLEY'S RECTIFICATION METHOD
	double H1[3][3], H2[3][3], iM[3][3];
	CvMat _H1 = cvMat(3, 3, CV_64F, H1);
	CvMat _H2 = cvMat(3, 3, CV_64F, H2);
	CvMat _iM = cvMat(3, 3, CV_64F, iM);

	cvStereoRectifyUncalibrated(&_imagePoints1, &_imagePoints2, &_F, imageSize, &_H1, &_H2, 3);
	cvInvert(&_M1, &_iM);
	cvMatMul(&_H1, &_M1, &_R1);
	cvMatMul(&_iM, &_R1, &_R1);
	cvInvert(&_M2, &_iM);
	cvMatMul(&_H2, &_M2, &_R2);
	cvMatMul(&_iM, &_R2, &_R2);

	//Precompute map for cvRemap
	cvReleaseMat(&mx1);
	cvReleaseMat(&my1);
	cvReleaseMat(&mx2);
	cvReleaseMat(&my2);
	mx1 = cvCreateMat(imageSize.height,imageSize.width, CV_32F);
	my1 = cvCreateMat(imageSize.height,imageSize.width, CV_32F);
	mx2 = cvCreateMat(imageSize.height,imageSize.width, CV_32F);
	my2 = cvCreateMat(imageSize.height,imageSize.width, CV_32F);

	cvInitUndistortRectifyMap(&_M1, &_D1, &_R1, &_M1, mx1, my1);
	cvInitUndistortRectifyMap(&_M2, &_D2, &_R2, &_M2, mx2, my2);
}

void StereoView::calibrateImages(IplImage *lImg, IplImage *rImg)
{
	if (ready)
	{
		// rectify left image
		lGrayTmpImage = cvCloneImage(lImg);
		cvRemap(lGrayTmpImage, lImg, mx1, my1);
		cvReleaseImage(&lGrayTmpImage);

		// rectify right image
		rGrayTmpImage = cvCloneImage(rImg);
		cvRemap(rGrayTmpImage, rImg, mx2, my2);
		cvReleaseImage(&rGrayTmpImage);
	}
}

